/*
 * File:        TFT, keypad, DAC, LED, PORT EXPANDER test
 *              With serial interface to PuTTY console
 *              DMA GetMachineBuffer
 *              !!!Modified scheduler!!!
 *              AUX uart loopback test
 *    -- Wire RB13 (U1RX) to RB7 (U1TX) for AUX machine UART1 loopback
 *       Runs up to 115200 baud, and probably 250000
 *    -- Wire RPA1  (U2RX) to USB TX line for console
 *    -- Wire RPB10 (U2TX) to USB RX line for console
 * 
 * Author:      Bruce Land
 * For use with Sean Carroll's Big Board
 * http://people.ece.cornell.edu/land/courses/ece4760/PIC32/target_board.html
 * Target PIC:  PIC32MX250F128B
 */

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config_1_3_3.h"
// threading library
#include "pt_cornell_1_3_3.h"
// yup, the expander
#include "port_expander_brl4.h"

////////////////////////////////////
// graphics libraries
// SPI channel 1 connections to TFT
#include "tft_master.h"
#include "tft_gfx.h"
// need for rand function
#include <stdlib.h>
// need for sin function
#include <math.h>
////////////////////////////////////

// lock out timer 2 interrupt during spi communication to port expander
// This is necessary if you use the SPI2 channel in an ISR.
// The ISR below runs the DAC using SPI2
#define start_spi2_critical_section INTEnable(INT_T2, 0)
#define end_spi2_critical_section INTEnable(INT_T2, 1)

////////////////////////////////////

/* Demo code for interfacing TFT (ILI9340 controller) to PIC32
 * The library has been modified from a similar Adafruit library
 */
// Adafruit data:
/***************************************************
  This is an example sketch for the Adafruit 2.2" SPI display.
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

// string buffer
//char buffer[60];

////////////////////////////////////
// DAC ISR
// A-channel, 1x, active
#define DAC_config_chan_A 0b0011000000000000
// B-channel, 1x, active
#define DAC_config_chan_B 0b1011000000000000
// DDS constant
#define two32 4294967296.0 // 2^32 
#define Fs 100000

//== Timer 2 interrupt handler ===========================================
// generates a DDS sinewave at a frequency set by the serial thread
// default frequency is 400 Hz.
//
volatile unsigned int DAC_data ;// output value
volatile SpiChannel spiChn = SPI_CHANNEL2 ;	// the SPI channel to use
volatile int spiClkDiv = 4 ; // 10 MHz max speed for port expander!!
// the DDS units:
volatile unsigned int phase_accum_main, phase_incr_main=100.0*two32/Fs ;//
// DDS sine table
#define sine_table_size 256
volatile int sin_table[sine_table_size];

void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
{
    int junk;
    
    mT2ClearIntFlag();
    
    // main DDS phase and sine table lookup
    phase_accum_main += phase_incr_main  ;
    DAC_data = sin_table[phase_accum_main>>24]  ;
 
    // === Channel A =============
    // wait for possible port expander transactions to complete
    while (TxBufFullSPI2());
    // reset spi mode to avoid conflict with port expander
    SPI_Mode16();
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
    // CS low to start transaction
     mPORTBClearBits(BIT_4); // start transaction
    // write to spi2 
    WriteSPI2( DAC_config_chan_A | ((DAC_data + 2048) & 0xfff));
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
     // CS high
    mPORTBSetBits(BIT_4); // end transaction
    // need to read SPI channel to avoid confusing port expander
    junk = ReadSPI2(); 
   //    
}

// === print a line on TFT =====================================================
// Utilities to print a line on the TFT
// Predefined colors definitions (from tft_master.h)
//#define	ILI9340_BLACK   0x0000
//#define	ILI9340_BLUE    0x001F
//#define	ILI9340_RED     0xF800
//#define	ILI9340_GREEN   0x07E0
//#define ILI9340_CYAN    0x07FF
//#define ILI9340_MAGENTA 0xF81F
//#define ILI9340_YELLOW  0xFFE0
//#define ILI9340_WHITE   0xFFFF

// string buffer
char buffer[60];
void printLine(int line_number, char* print_buffer, short text_color, short back_color){
    // line number 0 to 31 
    /// !!! assumes tft_setRotation(0);
    // print_buffer is the string to print
    int v_pos;
    v_pos = line_number * 10 ;
    // erase the pixels
    tft_fillRoundRect(0, v_pos, 239, 8, 1, back_color);// x,y,w,h,radius,color
    tft_setTextColor(text_color); 
    tft_setCursor(0, v_pos);
    tft_setTextSize(1);
    tft_writeString(print_buffer);
}

void printLine2(int line_number, char* print_buffer, short text_color, short back_color){
    // line number 0 to 31 
    /// !!! assumes tft_setRotation(0);
    // print_buffer is the string to print
    int v_pos;
    v_pos = line_number * 20 ;
    // erase the pixels
    tft_fillRoundRect(0, v_pos, 239, 16, 1, back_color);// x,y,w,h,radius,color
    tft_setTextColor(text_color); 
    tft_setCursor(0, v_pos);
    tft_setTextSize(2);
    tft_writeString(print_buffer);
}

// === Timer Thread =================================================
// system 1 second interval tick
// prints on TFT and blinks LED on RA0
int sys_time_seconds ;
// thread identifier to set thread parameters
int thread_num_timer;

// update a 1 second tick counter
static PT_THREAD (protothread_timer(struct pt *pt))
{
    PT_BEGIN(pt);
     // set up LED to blink
     mPORTASetBits(BIT_0 );	//Clear bits to ensure light is off.
     mPORTASetPinsDigitalOut(BIT_0 );    //Set port as output
      while(1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(1000) ;
        sys_time_seconds++ ;
        // toggle the LED on the big board
        mPORTAToggleBits(BIT_0);
        // draw sys_time
        sprintf(buffer,"Time=%d", sys_time_seconds);
        printLine2(0, buffer, ILI9340_BLACK, ILI9340_GREEN);
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // timer thread

// === Keypad Thread =============================================
// Decodes a keypad attached to the port expander, including a shift KEY.
//
// Port Expander connections:
// y0 -- row 1 -- thru 300 ohm resistor -- avoid short when two buttons pushed
// y1 -- row 2 -- thru 300 ohm resistor
// y2 -- row 3 -- thru 300 ohm resistor
// y3 -- row 4 -- thru 300 ohm resistor
// y4 -- col 1 -- internal pullup resistor -- avoid open circuit input when no button pushed
// y5 -- col 2 -- internal pullup resistor
// y6 -- col 3 -- internal pullup resistor
// y7 -- shift key connection -- internal pullup resistor

static PT_THREAD (protothread_key(struct pt *pt))
{
    PT_BEGIN(pt);
    static int keypad, i ;
    // order is 0 thru 9 then * ==10 and # ==11
    // no press = -1
    // table is decoded to natural digit order (except for * and #)
    // with shift key codes for each key
    // keys 0-9 return the digit number
    // keys 10 and 11 are * adn # respectively
    // Keys 12 to 21 are the shifted digits
    // keys 22 and 23 are shifted * and # respectively
    static int keytable[24]=
    //        0     1      2    3     4     5     6      7    8     9    10-*  11-#
            {0xd7, 0xbe, 0xde, 0xee, 0xbd, 0xdd, 0xed, 0xbb, 0xdb, 0xeb, 0xb7, 0xe7,
    //        s0     s1    s2  s3    s4    s5    s6     s7   s8    s9    s10-* s11-#
             0x57, 0x3e, 0x5e, 0x6e, 0x3d, 0x5d, 0x6d, 0x3b, 0x5b, 0x6b, 0x37, 0x67};
    // bit pattern for each row of the keypad scan -- active LOW
    // bit zero low is first entry
    static char out_table[4] = {0b1110, 0b1101, 0b1011, 0b0111};
    // init the port expander
    
    // Turn off the T2 ISR for a few microseconds
    start_spi2_critical_section;
    initPE();
    // PortY on Expander ports as digital outputs
    mPortYSetPinsOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);    //Set port as output
    // PortY as inputs
    // note that bit 7 will be shift key input, 
    // separate from keypad
    mPortYSetPinsIn(BIT_4 | BIT_5 | BIT_6 | BIT_7);    //Set port as input
    mPortYEnablePullUp(BIT_4 | BIT_5 | BIT_6 | BIT_7); 
    end_spi2_critical_section ;
    
    // the read-pattern if no button is pulled down by an output
    #define no_button (0x70)

      while(1) {
        // yield time for a keypad should be around 30 mSec
        PT_YIELD_TIME_msec(30);
        
        // stepping thru the 4 rows of the keypad
        for (i=0; i<4; i++) {
            
            start_spi2_critical_section ;
            // scan each row active-low
            writePE(GPIOY, out_table[i]);
            //reading the port also reads the outputs
            keypad  = readPE(GPIOY);
            end_spi2_critical_section ;
            
            // was there a keypress?
            if((keypad & no_button) != no_button) { break;}
        }
        
        // search for keycode
        if (keypad > 0){ // then button is pushed
            for (i=0; i<24; i++){
                if (keytable[i]==keypad) break;
            }
            // if invalid, two button push, set to -1
            if (i==24) i=-1;
        }
        else i = -1; // no button pushed

        // draw key number
        if (i>-1 && i<10) sprintf(buffer,"   %x %d", keypad, i);
        if (i==10 ) sprintf(buffer,"   %x *", keypad);
        if (i==11 ) sprintf(buffer,"   %x #", keypad);
        if (i>11 && i<22 ) sprintf(buffer, "   %x shift-%d", keypad, i-12);
        if (i==22 ) sprintf(buffer,"   %x ahift-*", keypad);
        if (i==23 ) sprintf(buffer,"   %x shift-#", keypad);
        if (i>-1 && i<12) printLine2(10, buffer, ILI9340_GREEN, ILI9340_BLACK);
        else if (i>-1) printLine2(10, buffer, ILI9340_RED, ILI9340_BLACK);
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // keypad thread

//=== Serial terminal thread =================================================
// simple command interpreter and serial display
// BUT NOTE that there are TWO completely different modes of getting strings
// from the UART. ONe assumes a human is typing, the other assumes a machine
//
// Also --  AUX uart loopback transmit thread using the 's' command

// semaphores to sync threads
int ready_to_send=1, ready_to_receive=0 ;
// NOTE that command 's' below sends/receives data via the 
// UART1 AUX machine communcation channel.
// while UART2 communicates with the console as usual

// mode==1 is command mode; 2 is data mode
static int mode = 1;

static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      static char cmd[30], t0;
      static float value;
      static int i;
      static int v1, v2;
      // string recognized by module to toggle data<->command
      static char cmd_data_toggle[] = "+++\r";
      // termination for machine buffer
      PT_terminate_char = '#' ; // see  http://www.asciitable.com/
      PT_terminate_count = 0 ;
      // time in milliseconds!
      PT_terminate_time = 0 ;
      // semaphores for sync two thread
            
      while(1) {  
        // === send string to UART2 (console) ==============
        // send the prompt via DMA to serial

        sprintf(PT_send_buffer,"\r\ncmd>");
        // by spawning a print thread
        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );

        // get interactive command
        PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer(&pt_input) );
        sscanf(PT_term_buffer, "%s %f", cmd, &value);

        // ==============================================
        // At this point, we have a valid string buffer
        // Now parse it:

        // === command parser ================================
         switch(cmd[0]){
            case 'f': // set frequency of DAC sawtooth output
                 // This command is here just to make sure the cpu is rnning
                 // enter frequecy in HZ
                phase_incr_main = (int)(value*(float)two32/Fs);
                sprintf(PT_send_buffer,"DDS freq = %6.1f", value);
                // by spawning a print thread
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );

                sprintf(PT_send_buffer," DDS increment = %d", phase_incr_main);
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                 break;

            case 'm':
                // toggle the bluetooth module mode between 'command' and 'data'
                // mode==1 is command mode
                // 
                if (mode == 1) {
                    // go to data mode
                    mode=2;
                    sprintf(PT_send_buffer,"Data mode");
                }
                else {
                    // goto command mode
                    mode = 1;
                    sprintf(PT_send_buffer,"Command mode");
                }
                // spawn a print thread for human
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                
                // now send the mode change to the module
                strcpy(PT_send_buffer_aux, cmd_data_toggle);
                PT_YIELD_UNTIL(pt, ready_to_send==1);                   
                ready_to_receive = 1;                  
                ready_to_send = 0 ;
                PT_YIELD(pt);
                PT_SPAWN(pt, &pt_DMA_output_aux, PT_DMA_PutSerialBuffer_aux(&pt_DMA_output_aux) );
                break;

            case 's':
                // read a string into the aux send buffer
                // IN COMMAND MODE:
                // -- returns bluetooth module response
                // IN DATA MODE:
                // -- returns string from remote bluetooth device
                // -- hangs until it gets a response with '\r' terminator
                //
                // get the string (command or payload data) to send to the module
                sscanf(PT_term_buffer, "%s %s", cmd, PT_send_buffer_aux);
                strcat(PT_send_buffer_aux, "\r");
                // 
                //send a string to bluetooth module thru UART1 (AUX))
                // wait for the read to be done
                PT_YIELD_UNTIL(pt, ready_to_send==1);
                // start the read, THEN start the write
                // signal the receive thread that data is coming
                ready_to_receive = 1;
                // clear the ready flag until read thread is done
                ready_to_send = 0 ;
                // wait a little so that receive thread is ready
                // and the DMA channel is ready
                PT_YIELD(pt);

                // send using AUX USART 
                // NOTE: set <CR> as ending character (of a string) in PT_DMA_PutSerialBuffer_aux
                // REQUIRED by ADAFRUIT SERIAL FRIEND
                // setting the end character to NULL transmits the NULL,
                // which confuses the module
                PT_SPAWN(pt, &pt_DMA_output_aux, PT_DMA_PutSerialBuffer_aux(&pt_DMA_output_aux) );

                // wait for the AUX receive to actually happen
                // (see receive thread))
                // (including time outs)
                PT_YIELD_UNTIL(pt, ready_to_send==1);

                // print response
                 memcpy(PT_send_buffer, PT_term_buffer_aux, max_chars_aux);
                // spawn a print thread to the terminal
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                break;
                
            case 'b':
                // send a burst of N data points to the module, WITHOUT waiting for a response
                // use  counter data generated for the dataa points.
                // MUST be in DATA mode!
                // get N = v1
                sscanf(PT_term_buffer, "%s %d", cmd, &v1);
                v2 = PT_GET_TIME();
                for (i=0; i<v1; i++) {
                    // format sample for send
                    sprintf(PT_send_buffer_aux, "%03d \r", i);
                    // send
                    PT_SPAWN(pt, &pt_DMA_output_aux, PT_DMA_PutSerialBuffer_aux(&pt_DMA_output_aux) );
                }
                sprintf(PT_send_buffer,"%d mSec\r\n", PT_GET_TIME()-v2);
                // spawn a print thread to the terminal
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                
                break;

         } // end switch statement

        // never exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread serial

// ========================================================
// AUX uart receive thread
static PT_THREAD (protothread_serial_aux(struct pt *pt))
{
    PT_BEGIN(pt);
      // termination for machine buffer on AUX channel
      // terminate on '#' OR 6 caracters OR 1000 mSec
      // whichever comes first
      PT_terminate_char_aux = 0 ; // see  http://www.asciitable.com/
      PT_terminate_count_aux = max_chars_aux ;
      // time in milliseconds!
      PT_terminate_time_aux = 500 ;
      
      while(1) {
          
        if (mode==1) { // command mode. No <CR> required, so can receive multiple lines
            PT_terminate_char_aux = 0 ; // allow multiple line receive
            PT_terminate_count_aux = max_chars_aux ;
            // time in milliseconds!
            PT_terminate_time_aux = 500 ;
            // wait for data transmission start
        }
        else { // data mode terminated by <CR>, no time limit
            PT_terminate_char_aux = '\r' ; // see  http://www.asciitable.com/
            PT_terminate_count_aux = max_chars_aux ;
            // time in milliseconds!
            PT_terminate_time_aux = 0 ; // no time out for radio respose
        }
        
        PT_YIELD_UNTIL(pt, ready_to_receive==1);
        ready_to_receive = 0 ;
        // get sent data on AUX channel (UART1 loopback)
        PT_SPAWN(pt, &pt_DMA_input_aux, PT_GetMachineBuffer_aux(&pt_DMA_input_aux) );  

        // reset semaphore to indicate data received
        // (or timed out)
        ready_to_send = 1;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // aux uart thread

// === Main  ======================================================

void main(void) {
  
  // === set up timer for DAC ISR =====================
  // timer interrupt //////////////////////////
    // Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
    // at 30 MHz PB clock 60 counts is two microsec
    // 400 is 100 ksamples/sec
    // 2000 is 20 ksamp/sec
    OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 400);

    // set up the timer interrupt with a priority of 2
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    mT2ClearIntFlag(); // and clear the interrupt flag
    
    // === SPI channel for DAC and Expander ============
    // control CS for DAC
    mPORTBSetPinsDigitalOut(BIT_4);
    mPORTBSetBits(BIT_4);
    // SCK2 is on pin RB15 
    // SDO2 (MOSI) is in PPS output group 2, and could be connected to RB5 which is pin 14
    PPSOutput(2, RPB5, SDO2);
    // 16 bit transfer CKP=1 CKE=1
    // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
    // For any given peripherial, you will need to match these
    // NOTE!! IF you are using the port expander THEN
    // >>> clk divider must be set to 4 for 10 MHz <<<
    SpiChnOpen(SPI_CHANNEL2, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV , 4);
  // end SPI setup
    
  // === build the DDS sine lookup table =======
   // scaled to produce values between 0 and 4096
   int ii;
   for (ii = 0; ii < sine_table_size; ii++){
         sin_table[ii] = (int)(2047*sin((float)ii*6.283/(float)sine_table_size));
    }
    
  // === setup system wide interrupts  ========
  INTEnableSystemMultiVectoredInt();
  
  // === TFT setup ============================
  // init the display in main since more than one thread uses it.
  // NOTE that this init assumes SPI channel 1 connections
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
  //240x320 vertical display
  tft_setRotation(0); // Use tft_setRotation(1) for 320x240
  
  // === config threads ========================
  PT_setup();
  
  // === identify the threads to the scheduler =====
  // add the thread function pointers to be scheduled
  // --- Two parameters: function_name and rate. ---
  // rate=0 fastest, rate=1 half, rate=2 quarter, rate=3 eighth, rate=4 sixteenth,
  // rate=5 or greater DISABLE thread!
  // If you need to access specific thread descriptors (to change rate), 
  // then return the list index
  
  thread_num_timer = pt_add(protothread_timer, 0);
  pt_add(protothread_key, 1);
  pt_add(protothread_serial, 1);
  pt_add(protothread_serial_aux, 0);

  // === initalize the scheduler ====================
  PT_INIT(&pt_sched) ;
  // >>> CHOOSE the scheduler method: <<<
  // (1)
  // SCHED_ROUND_ROBIN just cycles thru all defined threads
  //pt_sched_method = SCHED_ROUND_ROBIN ;
  
  // (2)
  // SCHED_RATE executes some threads more often then others
  // -- rate=0 fastest, rate=1 half, rate=2 quarter, rate=3 eighth, rate=4 sixteenth,
  // -- rate=5 or greater DISABLE thread!
  // pt_sched_method = SCHED_RATE ;
  
  pt_sched_method = SCHED_ROUND_ROBIN ;
  
  // === scheduler thread =======================
  // scheduler never exits
  PT_SCHEDULE(protothread_sched(&pt_sched));
  // ============================================
  
} // main

// === end  ======================================================

